//==============================================
//      File dbtreereg.pas
//
//      Components library for Delphi and BCB++.
//      DesignTime  TrDBxxxxTree-.
//      DesignTime for TrDBxxxxTree components.
//
//      Copyright 1998-2002 Polaris Software
//      http://polesoft.da.ru
//      mailto: PolarisLib@mail.ru
//==============================================
unit DBTreeReg;

{$I POLARIS.INC}

interface

uses Classes, TypInfo, SysUtils, DB, rDBTreeView,
     {$IFDEF POLARIS_DELPHI5}DBReg,{$ELSE}PolarisReg,{$ENDIF}
     {$IFDEF POLARIS_D6}DesignIntf,DesignEditors{$ELSE}DsgnIntf{$ENDIF};

procedure Register;

implementation

type
  TrTreeDataFieldProperty = class(TDBStringProperty)
  public
    procedure GetValueList(List: TStrings); override;
    function GetDataSourcePropName: string; virtual;
  end;

  TrTreeLookupFieldProperty = class(TDBStringProperty)
  public
    procedure GetValueList(List: TStrings); override;
    function GetDataSourcePropName: string; virtual;
  end;

function TrTreeDataFieldProperty.GetDataSourcePropName: string;
begin
//  Result := 'LookupSource';
  Result := 'DataSource';
end;

procedure TrTreeDataFieldProperty.GetValueList(List: TStrings);
var
  Instance: TComponent;
  PropInfo: PPropInfo;
  DataSource: TDataSource;
begin
  Instance := TComponent(GetComponent(0));
  PropInfo := TypInfo.GetPropInfo(Instance.ClassInfo, GetDataSourcePropName);
  if (PropInfo <> nil) and (PropInfo^.PropType^.Kind = tkClass) then
  begin
    DataSource := TObject(GetOrdProp(Instance, PropInfo)) as TDataSource;
    if (DataSource <> nil) and (DataSource.DataSet <> nil) then
      DataSource.DataSet.GetFieldNames(List);
  end;
end;

function TrTreeLookupFieldProperty.GetDataSourcePropName: string;
begin
  Result := 'LookupSource';
end;

procedure TrTreeLookupFieldProperty.GetValueList(List: TStrings);
var
  Instance: TComponent;
  PropInfo: PPropInfo;
  DataSource: TDataSource;
begin
  Instance := TComponent(GetComponent(0));
  PropInfo := TypInfo.GetPropInfo(Instance.ClassInfo, GetDataSourcePropName);
  if (PropInfo <> nil) and (PropInfo^.PropType^.Kind = tkClass) then
  begin
    DataSource := TObject(GetOrdProp(Instance, PropInfo)) as TDataSource;
    if (DataSource <> nil) and (DataSource.DataSet <> nil) then
      DataSource.DataSet.GetFieldNames(List);
  end;
end;

procedure Register;
begin
  RegisterComponents('Polaris DB', [TrDBTreeView, TrDBLookupTree, TrDBLookupComboTree]);

  RegisterPropertyEditor(TypeInfo(string), TrDBTreeView, 'KeyField',
    TrTreeDataFieldProperty);
  RegisterPropertyEditor(TypeInfo(string), TrDBTreeView, 'ChildrenField',
    TrTreeDataFieldProperty);
  RegisterPropertyEditor(TypeInfo(string), TrDBTreeView, 'ParentField',
    TrTreeDataFieldProperty);
  RegisterPropertyEditor(TypeInfo(string), TrDBTreeView, 'DisplayFields',
    TrTreeDataFieldProperty);

  RegisterPropertyEditor(TypeInfo(string), TrDBLookupTree, 'DataField',
    TrTreeDataFieldProperty);
  RegisterPropertyEditor(TypeInfo(string), TrDBLookupTree, 'LookupField',
    TrTreeLookupFieldProperty);
  RegisterPropertyEditor(TypeInfo(string), TrDBLookupTree, 'ChildrenField',
    TrTreeLookupFieldProperty);
  RegisterPropertyEditor(TypeInfo(string), TrDBLookupTree, 'ParentField',
    TrTreeLookupFieldProperty);
  RegisterPropertyEditor(TypeInfo(string), TrDBLookupTree, 'LookupDisplay',
    TrTreeLookupFieldProperty);

  RegisterPropertyEditor(TypeInfo(string), TrDBLookupComboTree, 'DataField',
    TrTreeDataFieldProperty);
  RegisterPropertyEditor(TypeInfo(string), TrDBLookupComboTree, 'LookupField',
    TrTreeLookupFieldProperty);
  RegisterPropertyEditor(TypeInfo(string), TrDBLookupComboTree, 'ChildrenField',
    TrTreeLookupFieldProperty);
  RegisterPropertyEditor(TypeInfo(string), TrDBLookupComboTree, 'ParentField',
    TrTreeLookupFieldProperty);
  RegisterPropertyEditor(TypeInfo(string), TrDBLookupComboTree, 'LookupDisplay',
    TrTreeLookupFieldProperty);
end;

end.
