//==============================================
//      File rrecgrid.pas
//
//      Components library for Delphi and BCB++.
//       .
//      rRecordGrid Component.
//
//      Copyright 1998-2000 Polaris Software
//      http://polesoft.da.ru
//      mailto: PolarisLib@mail.ru
//==============================================

unit rRecGrid;

{$I POLARIS.INC}
{$R-}

interface

uses SysUtils, Messages, Windows, Classes, Graphics, Menus, Controls, Forms,
     StdCtrls, rCRecGrd, DB, DBCtrls {$IFDEF POLARIS_D6}, Variants{$ENDIF};

{ TrRecordGrid }

type

  TValidateEvent = procedure(Sender: TObject; ARow: Longint; PickList: TStrings; var Value: Variant; var Accept: Boolean) of object;
  TGetTextEvent = procedure(Sender: TObject; ACol, ARow: Longint; PickList: TStrings; var Value: Variant) of object;
  TGetMaskEvent = procedure(Sender: TObject; ARow: Longint; var Mask: String) of object;
  TSetTextEvent = procedure(Sender: TObject; ARow: Longint; Modified: Boolean; PickList: TStrings; Const Value: Variant) of object;
  TSetEditEvent = procedure(Sender: TObject; ARow: Longint; PickList: TStrings; Const Value: Variant) of object;
  TMovedEvent = procedure(Sender: TObject; FromIndex, ToIndex: Longint) of object;
  TEditButtonClick = procedure(Sender: TObject; ARow: Longint; Field: TField) of object;

  TrRecordGrid = class;
  TRow = class;

  { TRowTitle }

  TRowTitle = class(TPersistent)
  private
    FColumn: TRow;
    FCaption: string;
    FFont: TFont;
    FColor: TColor;
    FAlignment: TAlignment;
    procedure FontChanged(Sender: TObject);
    function GetAlignment: TAlignment;
    function GetColor: TColor;
    function GetCaption: string;
    function GetFont: TFont;
    function IsAlignmentStored: Boolean;
    function IsColorStored: Boolean;
    function IsFontStored: Boolean;
    function IsCaptionStored: Boolean;
    procedure SetAlignment(Value: TAlignment);
    procedure SetColor(Value: TColor);
    procedure SetFont(Value: TFont);
    procedure SetCaption(const Value: string); virtual;
  protected
    procedure RefreshDefaultFont;
  public
    constructor Create(Column: TRow);
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    function DefaultAlignment: TAlignment;
    function DefaultColor: TColor;
    function DefaultFont: TFont;
    function DefaultCaption: string;
    procedure RestoreDefaults; virtual;
    property Column: TRow read FColumn;
  published
    property Alignment: TAlignment read GetAlignment write SetAlignment
      stored IsAlignmentStored;
    property Caption: string read GetCaption write SetCaption stored IsCaptionStored;
    property Color: TColor read GetColor write SetColor stored IsColorStored;
    property Font: TFont read GetFont write SetFont stored IsFontStored;
  end;

  { TrGridDataLink }

  TrGridDataLink = class(TDataLink)
  private
    FGrid: TrRecordGrid;
  protected
    procedure ActiveChanged; override;
    procedure CheckBrowseMode; override;
    procedure DataSetScrolled(Distance: Integer); override;
    procedure RecordChanged(Field: TField); override;
  public
    constructor Create(AGrid: TrRecordGrid);
    destructor Destroy; override;
  end;

  TColumnButtonStyle = (cbsAuto, cbsDropDown, cbsDropDownList, cbsEllipsis, cbsNone);

  TColumnValue = (cvAlignment, cvReadOnly, cvImeMode, cvImeName, cvFont, cvTitleCaption,
    cvTitleColor, cvTitleAlignment, cvTitleFont);
  TColumnValues = set of TColumnValue;
  TRecGridRows = class;

  { TRow }

  TRow = class(TCollectionItem)
  private
    FObject: TObject;
    FField: TField;
    FFieldName: String;
    FTitle: TRowTitle;
    FFont: TFont;
    FStored: Boolean;
    FValue: Variant;
    FPickList: TStrings;
    FPopupMenu: TPopupMenu;
    FDropDownRows: Cardinal;
    FButtonStyle: TColumnButtonStyle;
    FEditMask: String;
    FOnEditButtonClick: TNotifyEvent;
    FAlignment: TAlignment;
    FReadonly: Boolean;
    FAssignedValues: TColumnValues;
    FImeMode: TImeMode;
    FImeName: TImeName;

    FIndentLevel: Cardinal;
    FExpanded: Boolean;
    FNodes: TRecGridRows;
    FParent: TRow;
    FHasChildren: Boolean;
    FIDCode: String;

    FOnGetEditMask: TGetMaskEvent;
    FOnGetEditText: TGetTextEvent;
    FOnGetText: TGetTextEvent;
    FOnSetEditText: TSetEditEvent;
    FOnSetText: TSetTextEvent;
    FOnValidate: TValidateEvent;

    function GetImeMode: TImeMode;
    function GetImeName: TImeName;
    function IsImeModeStored: Boolean;
    function IsImeNameStored: Boolean;
    procedure SetImeMode(Value: TImeMode); virtual;
    procedure SetImeName(Value: TImeName); virtual;

    function GetAlignment: TAlignment;
    procedure SetAlignment(Value: TAlignment); virtual;
    function GetPickList: TStrings;
    function GetReadOnly: Boolean;
    function IsAlignmentStored: Boolean;
    function IsReadOnlyStored: Boolean;
    procedure SetButtonStyle(Value: TColumnButtonStyle);
    procedure SetPickList(Value: TStrings);
    procedure SetPopupMenu(Value: TPopupMenu);
    procedure SetReadOnly(Value: Boolean); virtual;
    procedure EditButtonClick;

    function GetField: TField;
    procedure SetField(Value: TField); virtual;
    procedure SetFieldName(Const Value: String);
    procedure SetTitle(Value: TRowTitle);
    //
    function  GetFont: TFont;
    procedure SetFont(Value: TFont);
    function IsFontStored: Boolean;
  protected
    function GetDisplayName: String; override;
    function GetEditMask: String;
    function GetText: String;
    procedure SetText(Value: String);
    procedure SetIndentLevel(Value: Cardinal);
    procedure SetExpanded(Value: Boolean);
    procedure SetIndex(Value: Integer); override;
    function SetParentAndLevel: Integer;
    function CreateTitle: TRowTitle; virtual;
    //
    procedure RefreshDefaultFont;
    procedure FontChanged(Sender: TObject);
    property IsStored: Boolean read FStored write FStored default True;
  public
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;
    function GetGrid: TrRecordGrid;
    function DefaultAlignment: TAlignment;
    function DefaultReadOnly: Boolean;
    function DefaultImeMode: TImeMode;
    function DefaultImeName: TImeName;
    procedure Assign(Source: TPersistent); override;
    procedure Refresh; Virtual;
    property Grid: TrRecordGrid read GetGrid;
    property Field: TField read GetField write SetField;
    property AssignedValues: TColumnValues read FAssignedValues;
    property Value: Variant read FValue write FValue;
    property Cargo: TObject read FObject write FObject;
    property Nodes: TRecGridRows read FNodes write FNodes;
    property Parent: TRow read FParent write FParent;
    property HasChildren: Boolean read FHasChildren Write FHasChildren;
    //
    function DefaultFont: TFont;
  published
    property PickList: TStrings read GetPickList write SetPickList;
    property FieldName: String read FFieldName write SetFieldName;
    property Title: TRowTitle read FTitle write SetTitle;
    property Alignment: TAlignment read GetAlignment write SetAlignment stored IsAlignmentStored;
    property ButtonStyle: TColumnButtonStyle read FButtonStyle write SetButtonStyle default cbsAuto;
    property DropDownRows: Cardinal read FDropDownRows write FDropDownRows default 7;
    property PopupMenu: TPopupMenu read FPopupMenu write SetPopupMenu;
    property ReadOnly: Boolean read GetReadOnly write SetReadOnly stored IsReadOnlyStored;
    property EditMask: String read GetEditMask write FEditMask;
    property ImeMode: TImeMode read GetImeMode write SetImeMode stored IsImeModeStored;
    property ImeName: TImeName read GetImeName write SetImeName stored IsImeNameStored;
    property OnEditButtonClick: TNotifyEvent read FOnEditButtonClick write FOnEditButtonClick;
    property OnGetEditMask: TGetMaskEvent read FOnGetEditMask write FOnGetEditMask;
    property OnGetEditText: TGetTextEvent read FOnGetEditText write FOnGetEditText;
    property OnGetText: TGetTextEvent read FOnGetText write FOnGetText;
    property OnSetEditText: TSetEditEvent read FOnSetEditText write FOnSetEditText;
    property OnSetText: TSetTextEvent read FOnSetText write FOnSetText;
    property OnValidate: TValidateEvent read FOnValidate write FOnValidate;
    property Text: String read GetText write SetText;
    property IndentLevel: Cardinal read FIndentLevel write SetIndentLevel default 0;
    property Expanded: Boolean read FExpanded write SetExpanded default False;
    property IDCode: String read FIDCode write FIDCode;
    //
    property Font: TFont read GetFont write SetFont stored IsFontStored;
  end;

  TColumnClass = class of TRow;

  { TRecGridRows }

  TRecGridRows = class(TCollection)
  private
    FGrid: TrRecordGrid;
//    FEmpty: Boolean;
    function GetColByID(Index: Variant): TRow;
    function GetColumn(Index: Variant): TRow;
    procedure SetColumn(Index: Variant; Value: TRow);
  protected
    function GetOwner: TPersistent; override;
  public
    constructor Create(Grid: TrRecordGrid; ColumnClass: TColumnClass; AddEmpty: Boolean);
    function Add: TRow;
    procedure Clear;
    procedure Delete(Index: Variant);
    procedure RebuildRows;
    property Grid: TrRecordGrid read FGrid;
    property Items[Index: Variant]: TRow read GetColumn write SetColumn; default;
  end;

  { TrRecordGrid }

  TRecGridState = (gsDefault, gsCustomized);

  TrRecordGrid = class(TrCustomRecGrid)
  private
    FDataLink: TrGridDataLink;
    FRows: TRecGridRows;
    FState: TRecGridState;
    FRebuildRowsIfEmpty: Boolean;
    FObject: TObject;
    FOnTopLeftChanged: TNotifyEvent;
    FOnColEnter: TNotifyEvent;
    FOnColExit: TNotifyEvent;
    FReadOnly: Boolean;
    FUpdating: Boolean;
    FOnEditButtonClick: TEditButtonClick;
    FUpdatingEditor: Boolean;
    FOriginalImeName: TImeName;
    FOriginalImeMode: TImeMode;
    procedure SetIme;
    procedure UpdateIme;
    procedure WMIMEStartComp(var Message: TMessage); message WM_IME_STARTCOMPOSITION;
    function GetDataSource: TDataSource;
    procedure SetDataSource(Value: TDataSource);
    procedure SetRecGridRows(Value: TRecGridRows);
    procedure WMSetFocus(var Message: TWMSetFocus); message WM_SetFOCUS;
    procedure WMKillFocus(var Message: TMessage); message WM_KillFocus;
    procedure SetState(const Value: TRecGridState);
  protected
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;//

    function GetEditLimit: Integer; override;

    procedure FocusCell(ACol, ARow: Longint; MoveAnchor: Boolean); override;
    procedure DrawCell(ACol, ARow: Longint; ARect: TRect;
        AState: TGridDrawState); override;
    function GetEditMask(ACol, ARow: Longint): String; override;
    function GetEditText(ACol, ARow: Longint): String; override;
    procedure SetEditText(ACol, ARow: Longint; const cValue: String); override;
    procedure TopLeftChanged; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    property FixedCols;
    function CreateEditor: TInplaceEdit; override;
    function CreateRows: TRecGridRows; dynamic;
    property EditRow;
    procedure ColEnter; override;
    procedure ColExit; override;
    property InUpdating: Boolean read FUpdating write FUpdating;
    procedure DblClick; override;
    procedure Loaded; override;
    procedure ResyncFields;
    function CanEditAcceptKey(Key: Char): Boolean; override;
    procedure ShowEditor; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
{$IFDEF POLARIS_D4}
    function ExecuteAction(Action: TBasicAction): Boolean; override;
    function UpdateAction(Action: TBasicAction): Boolean; override;
{$ENDIF}

    procedure DefaultHandler(var Msg); override;
    function CellRect(ACol, ARow: Longint): TRect;
    procedure MouseToCell(X, Y: Integer; var ACol, ARow: Longint);
    procedure Reset; override;
    procedure Refresh; Virtual;
    procedure RefreshData; Virtual;
    property DataLink: TrGridDataLink read FDataLink;
    property Cargo: TObject read FObject write FObject;
    property Canvas;
    property Row;
    property TabStops;
    property TopRow;
    property State: TRecGridState read FState write SetState;
    property RebuildRowsIfEmpty: Boolean read FRebuildRowsIfEmpty write FRebuildRowsIfEmpty;
  published
    property Rows: TRecGridRows read FRows write SetRecGridRows;
    property DataSource: TDataSource read GetDataSource write SetDataSource;
    property Anchors;
    property Constraints;
    property Color;
    property FixedRows;
    property BorderStyle;
    property DefaultRowHeight;
    property FixedColor;
    property Options;
    property ImeMode;
    property ImeName;
    property Align;
    property Ctl3D;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font;
    property GridLineWidth;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property VisibleRowCount;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
    property OnTopLeftChanged: TNotifyEvent read FOnTopLeftChanged write FOnTopLeftChanged;
    property OnColEnter: TNotifyEvent read FOnColEnter write FOnColEnter;
    property OnColExit: TNotifyEvent read FOnColExit write FOnColExit;
    property OnEditButtonClick: TEditButtonClick read FOnEditButtonClick write FOnEditButtonClick;
    property ReadOnly: Boolean Read FReadOnly write FReadOnly Default False;
  end;

  const ColumnTitleValues = [cvTitleColor..cvTitleFont];

implementation

type

  TEditStyle = (esSimple, esEllipsis, esPickList, esDropDown, esDataList);
  TPopupListbox = class;

  TrRecGridInplaceEdit = class(TInplaceEdit)
  private
    FButtonWidth: Integer;
    FDataList: TDBLookupListBox;
    FPickList: TPopupListbox;
    FActiveList: TWinControl;
    FLookupSource: TDatasource;
    FEditStyle: TEditStyle;
    FListVisible: Boolean;
    FTracking: Boolean;
    FPressed: Boolean;
    FSavedCancel: TDataSetNotifyEvent;
    procedure UpdateField;
    procedure ListMouseUp(Sender: TObject; Button: TMouseButton;
        Shift: TShiftState; X, Y: Integer);
    procedure SetEditStyle(Value: TEditStyle);
    procedure StopTracking;
    procedure TrackButton(X, Y: Integer);
    procedure CatchCancel(Dataset: TDataSet);
    procedure CMCancelMode(var Message: TCMCancelMode); message CM_CancelMode;
    procedure WMSetFocus(var Msg: TWMSetFocus); message WM_SETFOCUS;
    procedure WMCancelMode(var Message: TMessage); message WM_CancelMode;
    procedure WMKillFocus(var Message: TMessage); message WM_KillFocus;
    procedure WMLButtonDblClk(var Message: TWMLButtonDblClk); message wm_LButtonDblClk;
    procedure WMPaint(var Message: TWMPaint); message wm_Paint;
    procedure WMSetCursor(var Message: TWMSetCursor); message WM_SetCursor;
  protected
    procedure BoundsChanged; override;
    procedure DoDropDownKeys(var Key: Word; Shift: TShiftState);
    procedure DropDown;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
        X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
        X, Y: Integer); override;
    procedure PaintWindow(DC: HDC); override;
    procedure UpdateContents; override;
    procedure WndProc(var Message: TMessage); override;
    function GetColumn: TRow;
    property EditStyle: TEditStyle read FEditStyle write SetEditStyle;
    property ActiveList: TWinControl read FActiveList write FActiveList;
    property DataList: TDBLookupListBox read FDataList;
    property PickList: TPopupListbox read FPickList;
//
//        procedure SetFocus; override;
//
  public
    procedure CloseUp(Accept: Boolean);
    constructor Create(Owner: TComponent); override;
  end;

{ TPopupListbox }

  TPopupListbox = class(TCustomListbox)
  private
    FSearchText: String;
    FSearchTickCount: Longint;
  protected
    procedure CreateParams(var Params: TCreateParams); override;
    procedure CreateWnd; override;
    procedure KeyPress(var Key: Char); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
  end;

procedure KillMessage(Wnd: HWnd; Msg: Integer);
// Delete the requested message from the queue, but throw back
// any WM_QUIT msgs that PeekMessage may also return
var
  M: TMsg;
begin
  M.Message := 0;
  if PeekMessage(M, Wnd, Msg, Msg, pm_Remove) and (M.Message = WM_QUIT) then
      PostQuitMessage(M.wparam);
end;

function GetSpecialText(Field: TField): String;
begin
  case Field.Datatype Of
    ftBytes,
    fttypedBinary,
    ftvarBytes: Result := '<Binary>';
    ftBlob    : Result := '<Blob>';
    ftMemo    : Result := Field.AsString;// '<Memo>';
    ftGraphic : Result := '<Bitmap>';
    ftFmtMemo : Result := '<Formatted>';
    ftParadoxOle,
    ftDBaseOle: Result := '<OLE>';
  else
    Result := '<>';
  end;
end;

function IsSpecialField(Field: TField): Boolean;
begin
  Result := Field.Datatype In [ftUnknown, ftBytes, fttypedBinary, ftvarBytes,
      ftBlob, ftMemo, ftGraphic, ftFmtMemo, ftParadoxOle, ftDBaseOle];
end;

{ TrRecordGrid }

procedure TrRecordGrid.ResyncFields;
var
  i: Integer;
begin
  for i := 0 to Rows.Count - 1 do
    if FDatalink.Active
      then Rows[i].FieldName := Rows[i].FieldName
      else Rows[i].FField := nil;
end;

procedure TrRecordGrid.ShowEditor;
begin
  if Rows[Row].Field <> nil
  then inherited ShowEditor;
end;

procedure TrRecordGrid.Refresh;
begin
  UpdateEdit;
  Invalidate;
end;

function TrRecordGrid.CanEditAcceptKey(Key: Char): Boolean;
begin
  //Rabbit (  :)
  if (EditRow >= 0) and (EditRow < Rows.Count) then
    with Rows[EditRow] do
      Result := FDatalink.Active and Assigned(Field) and Field.IsValidChar(Key)
  else Result := False;
end;

procedure TrRecordGrid.RefreshData;
begin
  UpdateEdit;
  InvalidateData;
end;

function TrRecordGrid.CellRect(ACol, ARow: Longint): TRect;
begin
  Result := inherited CellRect(ACol, ARow);
end;

procedure TrRecordGrid.MouseToCell(X, Y: Integer; var ACol, ARow: Longint);
var
  Coord: TGridCoord;
begin
  Coord := MouseCoord(X, Y);
  ACol := Coord.X;
  ARow := Coord.Y;
end;

function TrRecordGrid.GetEditMask(ACol, ARow: Longint): String;
begin
  Result := Rows[ARow].GetEditMask;
end;

function TrRecordGrid.GetEditText(ACol, ARow: Longint): String;
var VResult: Variant;
begin
  Result := '';
  if ARow < Rows.Count then begin
    with Rows[ARow] do  begin
      if Assigned(Field) then begin
        if DataSource.DataSet.Active then begin
          if IsSpecialField(Field)
            then Result := GetSpecialText(Field)
            else Result := DataSource.DataSet.FieldByName(Rows[ARow].FieldName).Text;
        end;
      end else
        if not (VarType(Value) = varEmpty)
          then Result := Value;
      if Assigned(OnGetEditText) then begin
        VResult := Result;
        OnGetEditText(Self, ACol, ARow, PickList, VResult);
        Result := VResult;
      end;
    end;
  end;
end;

procedure TrRecordGrid.SetEditText(ACol, ARow: Longint; Const cValue: String);
begin
  if ARow < Rows.Count then begin
    with Rows[ARow] do begin
      if {(Rows[ARow] <> nil) and} Assigned(Field) then begin
        if DataSource.DataSet.Active then begin
          if Assigned(Field) and
             (cValue <> Field.Text)
             and not IsSpecialField(Field)
             and (EditRow  =ARow) then
            if not Self.ReadOnly
               and not Rows[ARow].ReadOnly
               and not Field.ReadOnly
               and not (DataSource.DataSet.State In [dsEdit, dsInsert]) then begin
              InUpdating := True;
              DataSource.DataSet.Edit;
              Field.Text := cValue;
            end;
//          if (cValue <> Field.Text) and not IsSpecialField(Field) and (EditRow=ARow) then begin
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//          if (cValue <> Field.AsString) and not IsSpecialField(Field) then begin

//            if not Rows[ARow].ReadOnly and not Field.ReadOnly and not (DataSource.DataSet.State In [dsEdit, dsInsert]) then begin
//            if {not Self.ReadOnly and} not Rows[ARow].ReadOnly and not Field.ReadOnly and not (DataSource.DataSet.State In [dsEdit, dsInsert]) then begin
//              InUpdating := True;
//              DataSource.DataSet.Edit;
//            end;
//            Field.Text := cValue;
//          end;
        end;
      end else
        if not (vartype(Value) = varEmpty)
          then Value := cValue;
      if Assigned(OnSetEditText)
        then OnSetEditText(Self, ARow, PickList, cValue);
    end;
  end;
end;

procedure TrRecordGrid.TopLeftChanged;
begin
  inherited TopLeftChanged;
  if Assigned(FOnTopLeftChanged) then FOnTopLeftChanged(Self);
end;

function TrRecordGrid.GetDataSource: TDataSource;
begin
  Result := FDataLink.DataSource;
end;

procedure TrRecordGrid.SetDataSource(Value: TDataSource);
var
  nCnt: Integer;
  nSub: Integer;

  function CheckDataSet(Value: TDataSource): Boolean;
  begin
    Result := (Value<>nil) and (Value.DataSet<>nil);
  end;

begin
  if Value = FDatalink.Datasource then Exit;
//  if FDataLink.DataSource <> nil then begin
//    FDataLink.CheckBrowseMode;
//    if FDataLink.DataSource.DataSet.State in [dsEdit,dsInsert]
//      then FDataLink.DataSource.DataSet.Post;
//  end;
  FDataLink.DataSource := Value;
  if Value<>nil then Value.FreeNotification(Self);
  for nCnt := Rows.Count - 1 downTo 0 do begin
    with Rows[nCnt].Nodes do
      for nSub := Count -1 downTo 0 do
        if Assigned(TRow(Items[nSub]).Field) then
          if not CheckDataSet(Value)
            and (Value.DataSet.FindField(TRow(Items[nSub]).Field.FieldName) = nil)
            then Items[nSub].Collection := nil
            else TRow(Items[nSub]).Field := Value.DataSet.FindField(TRow(Items[nSub]).Field.FieldName);
      if Assigned(Rows[nCnt].Field) then
        if not CheckDataSet(Value)
          and (Value.DataSet.FindField(Rows[nCnt].Field.FieldName) = nil)
          then Rows[nCnt].Collection := nil
          else Rows[nCnt].Field := Value.DataSet.FindField(Rows[nCnt].Field.FieldName);
  end;
  RowCount := Rows.Count;
  if Row > RowCount - 1 then Row := RowCount - 1;
  Invalidate;
end;

procedure TrRecordGrid.SetRecGridRows(Value: TRecGridRows);
begin
  Rows.Assign(Value);
end;

function TrRecordGrid.CreateRows: TRecGridRows;
begin
  Result := TRecGridRows.Create(Self, TRow, False{True});
end;

procedure TrRecordGrid.SetIme;
var
  Column: TRow;
begin
  if not SysLocale.Fareast then Exit;
  if FUpdatingEditor then
  begin
    ImeName := Screen.DefaultIme;
    ImeMode := imDontCare;
  end
  else
  begin
    Column := Rows[Row];
    ImeName := FOriginalImeName;
    ImeMode := FOriginalImeMode;
    if cvImeMode In Column.FAssignedValues then
    begin
      ImeName := Column.ImeName;
      ImeMode := Column.ImeMode;
    end;
  end;
  if InplaceEditor <> nil then
  begin
    TrRecGridInplaceEdit(Self).ImeName := ImeName;
    TrRecGridInplaceEdit(Self).ImeMode := ImeMode;
  end;
end;

procedure TrRecordGrid.UpdateIme;
begin
  if not SysLocale.Fareast then Exit;
  SetIme;
  if InplaceEditor <> nil then
      TrRecGridInplaceEdit(Self).SetIme;
end;

procedure TrRecordGrid.WMIMEStartComp(var Message: TMessage);
begin
  inherited;
  FUpdatingEditor := True;
  ShowEditor;
  FUpdatingEditor := False;
end;

function TrRecordGrid.CreateEditor: TInplaceEdit;
begin
  Result := TrRecGridInplaceEdit.Create(Self);
end;

procedure TrRecordGrid.Notification(AComponent: TComponent; Operation: TOperation);
var
  I: Integer;
  nSub: Integer;
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) then begin
    if (AComponent Is TPopupMenu) then  begin
      for I := 0 to Rows.Count - 1 do begin
        if Rows[I].PopupMenu = AComponent
          then Rows[I].PopupMenu := nil;
        if Rows[I].Nodes.Count > 0 then
          for nSub := 0 to Rows[I].Nodes.Count - 1 do
            if TRow(Rows[I].Nodes.Items[nSub]).PopupMenu = AComponent
              then TRow(Rows[I].Nodes[nSub]).PopupMenu := nil;
      end;
    end else
    if (FDataLink <> nil) then
      if (AComponent = DataSource)
        then DataSource := nil
        else if (AComponent Is TField) then begin
          for I := 0 to Rows.Count - 1 do
            with Rows[I] do
              if Field = AComponent then begin
                Field := nil;
                if Nodes.Count > 0 then
                  for nSub := 0 to Nodes.Count - 1 do
                    if TRow(Nodes.Items[nSub]).Field = AComponent
                      then Field := nil;
              end;
        end;
  end;
end;

procedure TrRecordGrid.DrawCell(ACol, ARow: Longint; ARect: TRect; AState: TGridDrawState);

  procedure WriteText(ACanvas: TCanvas; ARect: TRect; DX, DY: Integer; Const Text: String; Alignment: TAlignment);
  const AlignFlags: Array[TAlignment] Of Integer =
          (DT_LEFT Or DT_WORDBREAK Or DT_EXPANDTABS Or DT_NOPREFIX,
          DT_RIGHT Or DT_WORDBREAK Or DT_EXPANDTABS Or DT_NOPREFIX,
          DT_CENTER Or DT_WORDBREAK Or DT_EXPANDTABS Or DT_NOPREFIX);
  var Left: Integer;
  begin
    case Alignment Of
      taLeftJustify:  Left := ARect.Left + DX;
      taRightJustify: Left := ARect.Right - ACanvas.TextWidth(Text) - 3;
    else // taCenter
      Left := ARect.Left + (ARect.Right - ARect.Left) Shr 1 - (ACanvas.TextWidth(Text) Shr 1);
    end;
    ExtTextOut(ACanvas.Handle, Left, ARect.Top + DY, ETO_OPAQUE Or ETO_CLIPPED, @ARect, PChar(Text), Length(Text), nil);
  end;

var
  S: String;
  Alignment: TAlignment;
  Val: Variant;
  DE: Boolean;
  DefaultRow : TRow;
begin
  S := '';
  DE := False;
  Alignment := taLeftJustify;
  if Rows.Count > ARow then begin
    if ACol = 0 then begin
      S := '  ';
      if Rows[ARow].HasChildren then begin
        if Rows[ARow].IndentLevel > 0
          then S := StringOfChar(' ', (Rows[ARow].IndentLevel) * 2)
          else S := '';
        if Rows[ARow].Expanded
          then S := S + '- '
          else S := S + '+';
      end else
        if Rows[ARow].IndentLevel > 0
          then S := StringOfChar(' ', (Rows[ARow].IndentLevel) * 2) + S;
      if Rows[ARow].Title = nil
      then S := S
      else S := S + Rows[ARow].Title.Caption;
            Canvas.Brush.Color := Rows[ARow].Title.Color;
      Canvas.FillRect(ARect);
      Canvas.Font.Assign(Rows[ARow].Title.Font);
      WriteText(Canvas, ARect, 2, 2, S, Rows[ARow].Title.Alignment);
      Exit;

    end else if Assigned(DataSource) and Assigned(DataSource.DataSet) then begin
      DefaultRow := Rows[ARow];
      Alignment := DefaultRow.Alignment;
      Canvas.Font.Assign(DefaultRow.Font);
      if (gdSelected in AState) and (gdFocused in AState) then
        Canvas.Font.Color := clHighlightText;
      with DefaultRow do begin
        if Assigned(Field) then begin
          if IsSpecialField(Field)
            then S := GetSpecialText(Field)
            else
              if DataSource.DataSet.Active
                then S := DataSource.DataSet.FieldByName(DefaultRow.FieldName).DisplayText;
        end
        else begin
          if not (VarType(Value) = varEmpty) then S := Value;
          DE := True
        end;
        if Assigned(OnGetText) then begin
          Val := S;
          OnGetText(Self, ACol, ARow, PickList, Val);
          S := Val;
        end;
      end;
    end;
  end;
  if DE
  then begin
    Canvas.Brush.Color := FixedColor;
    Canvas.FillRect(ARect);
    Canvas.Font.Color := clWindowText;
    WriteText(Canvas, ARect, 2, 2, S, Alignment);
    {
    TR := ARect;
    if gdFocused in AState
    then begin
      TR.Bottom := ARect.Bottom-1;
      TR.Right := ARect.Right-1;
    end
    else begin
      TR.Bottom := ARect.Bottom+1;
      TR.Right := ARect.Right;
    end;
    DrawEdge(Canvas.Handle, TR, DrwEdge[gdFocused in AState], BF_RECT);
    }
  end
  else WriteText(Canvas, ARect, 2, 2, S, Alignment);
end;

procedure TrRecordGrid.ColEnter;
begin
  UpdateIme;
  if Assigned(FOnColEnter) then FOnColEnter(Self);
end;

procedure TrRecordGrid.ColExit;
begin
  if Assigned(FOnColExit) then FOnColExit(Self);
end;

procedure TrRecordGrid.WMSetFocus(var Message: TWMSetFocus);
begin
  SetIme;
  inherited;
end;

procedure TrRecordGrid.WMKillFocus(var Message: TMessage);
begin
  ImeName := Screen.DefaultIme;
  ImeMode := imDontCare;
  inherited;
end;

procedure TrRecordGrid.DefaultHandler(var Msg);
var
  P: TPopupMenu;
  Cell: TGridCoord;
begin
  inherited DefaultHandler(Msg);
  if TMessage(Msg).Msg = wm_RButtonUp then
    with TWMRButtonUp(Msg) do
    begin
        Cell := MouseCoord(XPos, YPos);
        if (Cell.X <> -1) and (Cell.Y <> -1) then
            P := Rows[Cell.Y].PopupMenu
        else
            P := nil;
        if (P <> nil) and P.AutoPopup then
        begin
            SendCancelMode(nil);
            P.PopupComponent := Self;
            with ClientToScreen(SmallPointToPoint(Pos)) do
                P.Popup(X, Y);
            Result := 1;
        end;
    end;
end;

procedure TrRecordGrid.DblClick;
begin
  if Rows[Row].HasChildren then Rows[Row].Expanded := not Rows[Row].Expanded;
  inherited DblClick;
end;

procedure TrRecordGrid.Loaded;
var
  nCnt: Integer;
begin
  inherited Loaded;
  for nCnt := Rows.Count - 1 downto 0 do
  begin
    if not Rows[nCnt].Expanded then
    begin
      Rows[nCnt].FExpanded := True;
      Rows[nCnt].Expanded := False;
    end;
  end;
end;

constructor TrRecordGrid.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FRows := CreateRows;
  FDataLink := TrGridDataLink.Create(Self);
  FOriginalImeName := ImeName;
  FOriginalImeMode := ImeMode;
  FObject := nil;
  FRebuildRowsIfEmpty := True;
end;

procedure TrRecordGrid.Reset;
begin
  FRows.Free;
  FRows := nil;
  FRows := CreateRows;
  EditRow := 0;
  RowCount := 0;
  inherited Reset;
  Invalidate;
end;

destructor TrRecordGrid.Destroy;
begin
  FRows.Free;
  FRows := nil;
  FDataLink.Free;
  FDataLink := nil;
  inherited Destroy;
end;

{$IFDEF POLARIS_D4}
function TrRecordGrid.ExecuteAction(Action: TBasicAction): Boolean;
begin
  Result := inherited ExecuteAction(Action) or (FDataLink <> nil) and
    FDataLink.ExecuteAction(Action);
end;

function TrRecordGrid.UpdateAction(Action: TBasicAction): Boolean;
begin
  Result := inherited UpdateAction(Action) or (FDataLink <> nil) and
    FDataLink.UpdateAction(Action);
end;
{$ENDIF}

function TrRecordGrid.GetEditLimit: Integer;
begin
  Result := 0;
  if {(EditRow<>ARow) and }
      Assigned(Rows[EditRow]) and //RSV new
      (Rows[EditRow].Field <> nil) and
     (Rows[EditRow].Field.Datatype = ftString)
  then Result := Rows[EditRow].Field.Size;
end;

//!!!!!!!!!!!

procedure TrRecordGrid.FocusCell(ACol, ARow: Longint; MoveAnchor: Boolean);
begin
//  if (EditRow<>ARow) and (Rows[ARow].Field <> nil)
//  THEN Editor.Text := Rows[ARow].Field.Text;
  inherited FocusCell(ACol, ARow, MoveAnchor);
//  MoveCurrent(ACol, ARow, MoveAnchor, True);
//  UpdateEdit;
//  Click;
end;

procedure TrRecordGrid.SetState(const Value: TRecGridState);
begin
  if FState <> Value then FState := Value;
end;

procedure TrRecordGrid.KeyDown(var Key: Word; Shift: TShiftState);
begin
  if (ssCtrl in Shift) and Rows[Row].HasChildren
  then case Key of
  VK_ADD:      Rows[Row].Expanded := True;
  VK_SUBTRACT: Rows[Row].Expanded := False;
  end;
  if not (ssShift in Shift) and (Key=VK_ESCAPE)
  then begin
    if not (goAlwaysShowEditor in Options) then HideEditor;
    if (Perform(CM_WANTSPECIALKEY, DLGC_WANTALLKEYS, 0)=0) and
       (Perform(WM_GETDLGCODE, 0, 0) and DLGC_WANTALLKEYS = 0)
    then GetParentForm(Self).Perform(CM_DIALOGKEY, VK_ESCAPE, 0);
  end
  else inherited KeyDown(Key, Shift);
end;

{ TRow }

procedure TRow.SetIndentLevel(Value: Cardinal);
var
  nPar: Integer;
begin
  if Index = 0 then
    Value := 0
  else
    if Value > TRow(Collection.Items[Index - 1]).IndentLevel + 1 then
        Value := TRow(Collection.Items[Index - 1]).IndentLevel + 1;
  if FIndentLevel <> Value then
  begin
    FIndentLevel := Value;
    if Value > 0 then
    begin
      nPar := Index;
      repeat
        Dec(nPar)
      until TRow(Collection.Items[nPar]).IndentLevel < IndentLevel;
      Parent := TRow(Collection.Items[nPar]);
      Parent.HasChildren := True;
    end
    else
    begin
      if Index > 0 then
        TRow(Collection.Items[Index - 1]).HasChildren :=
            (TRow(Collection.Items[Index - 1]).IndentLevel < IndentLevel);
      Parent := nil;
      if Index < Collection.Count - 1 then
        HasChildren := (TRow(Collection.Items[Index + 1]).IndentLevel > IndentLevel)
      else
        HasChildren := False;
    end;
    Grid.Invalidate;
  end;
end;

function TRow.SetParentAndLevel: Integer;
var
  Idx: Integer;
  nCnt: Integer;
begin
  Idx := Index;
  Collection.beginUpdate;
  if Idx < Collection.Count then
  begin
    with Collection do
      repeat
        if TRow(Items[Idx]).Nodes.Count > 0 then
        begin
          TRow(Items[Idx]).HasChildren := True;
          for nCnt := 0 to Count - 1 do
          begin
            with TRow(TRow(Items[Idx]).Nodes.Items[nCnt]) do
            begin
              IndentLevel := TRow(Items[Idx]).IndentLevel + 1;
              Parent := TRow(Items[Idx]);
            end;
          end;
        end
        else
          TRow(Items[Idx]).HasChildren := False;
        if (Idx < Count - 1) and TRow(Items[Idx]).HasChildren then
          if TRow(Items[Idx + 1]).IndentLevel > IndentLevel then
          begin
              TRow(Items[Idx]).Expanded := not TRow(Items[Idx]).Expanded;
              TRow(Items[Idx]).Expanded := not TRow(Items[Idx]).Expanded;
          end;
        if TRow(Items[Idx]).IndentLevel > IndentLevel then
        begin
          if TRow(Items[Idx]).IndentLevel > IndentLevel + 1 then
              TRow(Items[Idx]).IndentLevel := IndentLevel + 1;
          TRow(Items[Idx]).Parent := TRow(Items[Idx - 1]);
          TRow(Items[Idx]).Parent.HasChildren := True;
          Idx := TRow(Items[Idx]).SetParentAndLevel;
        end
        else
        begin
          if TRow(Items[Idx]).IndentLevel = IndentLevel then
          begin
            TRow(Items[Idx]).Parent := Parent;
            Inc(Idx);
          end
          else
          begin
            Break;
          end;
        end;
      until Idx = Count;
  end;
  Collection.EndUpdate;
  Result := Idx;
end;

procedure TRow.SetIndex(Value: Integer);
begin
  inherited SetIndex(Value);
  if csDesigning In Grid.ComponentState then
    with Grid do
    begin
      if Value = 0 then
          if IndentLevel > 0 then IndentLevel := 0;
      Rows[0].Parent := nil;
      Rows[0].SetParentAndLevel;
    end;
  Grid.Invalidate;
end;

procedure TRow.SetExpanded(Value: Boolean);
var
  nIdx: Integer;
  nLvl: Cardinal;
begin
  { Always expanded while designing }
  if Value <> FExpanded then
  begin
    FExpanded := Value;
    if Value then // Expand
    begin
      if Nodes.Count > 0 then
      begin
        nIdx := Index;
        Collection.BeginUpdate;
        repeat
          Inc(nIdx);
          Nodes.Items[0].Collection := Self.Collection;
          Collection.Items[Collection.Count - 1].Index := nIdx;
        until Nodes.Count = 0;
//        Grid.ResyncFields;
        Collection.EndUpdate;
      end;
    end
    else // Collapse
    begin
      nIdx := Index;
      nLvl := FIndentLevel + 1;
      if not (csDesigning in Grid.ComponentState) then
      begin
        if Index < Collection.Count - 1 then
        begin
          Inc(nIdx);
          Collection.beginUpdate;
          while nIdx < Collection.Count do
          begin
            if TRow(Collection.Items[nIdx]).IndentLevel < nLvl then
              Break
            else
            begin
              { Forward Collapse }
              if nIdx + 1 < Collection.Count then
                if TRow(Collection.Items[nIdx + 1]).IndentLevel > nLvl then
                    TRow(Collection.Items[nIdx]).Expanded := False;
              { Collapse }
              Collection.Items[nIdx].Collection := Self.Nodes;
            end;
          end;
          Collection.EndUpdate;
        end;
      end
      {else
      begin
        if Index < Collection.Count - 1 then
          repeat
            Inc(nIdx);
            if TRow(Collection.Items[nIdx]).HasChildren then
              TRow(Collection.Items[nIdx]).Expanded := False;
            if TRow(Collection.Items[nIdx]).IndentLevel < nLvl then Break
          until nIdx = Collection.Count - 1;
      end;}
    end;
    GetGrid.RowCount := Collection.Count;
    Grid.Invalidate;
  end;
end;

function TRow.GetField: TField;
var
  Grid: TrRecordGrid;
begin
  // Returns nil if FieldName can't be found in dataset
  Grid := GetGrid;
  if (FField = nil) and (Length(FFieldName) > 0) and Assigned(Grid)
  and Assigned(Grid.DataLink.DataSet) then
    with Grid.Datalink.Dataset do
      if Active
        then SetField(FindField(FieldName));
  Result := FField;
end;

procedure TRow.SetField(Value: TField);
begin
  if FField = Value then Exit;
  FField := Value;
  if Assigned(Value)
    then  FFieldName := Value.FieldName;

  if Assigned(Value) and (FieldName <> Value.FieldName) then begin
    if (AnsiUpperCase(IDCode) = AnsiUpperCase(FFieldName)) or
      ((Title <> nil) and (AnsiUpperCase(IDCode) = AnsiUpperCase(Title.Caption))) or (IDCode = '')
        then FIDCode := Value.FieldName;
    if (Title <> nil) and ((AnsiUpperCase(Title.Caption) = AnsiUpperCase(FFieldName)) or (Title.Caption = ''))
      then FTitle.Caption := Value.FieldName;
  end;

  Changed(False);
  GetGrid.InvalidateRow(Index);
end;

procedure TRow.SetFieldName(const Value: String);
var
  AField: TField;
  Grid: TrRecordGrid;
  EmptyVar: Variant;
begin
  AField := nil;
  Grid := GetGrid;
  if Assigned(Grid) and Assigned(Grid.DataLink.DataSet) and
  not (csLoading In Grid.ComponentState) and (Length(Value) > 0)
    then AField := Grid.DataLink.DataSet.FindField(Value); { no exceptions }
  FFieldName := Value;
  SetField(AField);
  if AField <> nil then begin
    FTitle.Caption := AField.DisplayLabel;
    Self.Value := EmptyVar;
  end;
  GetGrid.InvalidateRow(Index);
  Changed(False);
end;

function TRow.GetGrid: TrRecordGrid;
begin
  if Assigned(Collection) and (Collection Is TRecGridRows)
    then Result := TRecGridRows(Collection).Grid
    else Result := nil;
end;

function TRow.GetDisplayName: String;
begin
  if (Title <> nil) then begin
    if (Title.Caption = EmptyStr) then begin
      if FFieldName = EmptyStr
        then Result := 'Column'
        else Result := FFieldName;
    end else
      Result := Title.Caption;
  end;
  Result := StringOfChar(' ',IndentLevel* 2) + Result;
end;

procedure TRow.Assign(Source: TPersistent);
begin
  if Source Is TRow then begin
    if Assigned(Collection) then Collection.beginUpdate;
    try
      FieldName := TRow(Source).FieldName;

      if cvAlignment In TRow(Source).AssignedValues then
          Alignment := TRow(Source).Alignment;
      if cvReadOnly In TRow(Source).AssignedValues then
          ReadOnly := TRow(Source).ReadOnly;
      if cvImeMode In TRow(Source).AssignedValues then
          ImeMode := TRow(Source).ImeMode;
      if cvImeName In TRow(Source).AssignedValues then
          ImeName := TRow(Source).ImeName;

      if Title <> nil then Title.Caption := TRow(Source).Title.Caption;
      DropDownRows := TRow(Source).DropDownRows;
      ButtonStyle := TRow(Source).ButtonStyle;
      PickList := TRow(Source).PickList;
      PopupMenu := TRow(Source).PopupMenu;
    finally
        if Assigned(Collection) then Collection.EndUpdate;
    end;
  end else
    inherited Assign(Source);
end;


function TRow.DefaultAlignment: TAlignment;
begin
  if Assigned(Field)
    then Result := Field.Alignment
    else Result := taLeftJustify;
end;

function TRow.DefaultReadOnly: Boolean;
var
  Grid: TrRecordGrid;
begin
  Grid := GetGrid;
  Result := (Assigned(Grid) and Grid.ReadOnly) or (Assigned(Field) and FField.ReadOnly);
end;

function TRow.GetAlignment: TAlignment;
begin
  if cvAlignment In FAssignedValues
    then Result := FAlignment
    else Result := DefaultAlignment;
end;

function TRow.GetPickList: TStrings;
begin
  if FPickList = nil then
    FPickList := TStringList.Create;
  Result := FPickList;
end;

function TRow.GetReadOnly: Boolean;
begin
  if cvReadOnly In FAssignedValues then
      Result := FReadOnly
  else
      Result := DefaultReadOnly;
end;

function TRow.IsAlignmentStored: Boolean;
begin
  Result := (cvAlignment In FAssignedValues) and (FAlignment <> DefaultAlignment);
end;

function TRow.IsReadOnlyStored: Boolean;
begin
  Result := (cvReadOnly In FAssignedValues) and (FReadOnly <> DefaultReadOnly);
end;

procedure TRow.SetAlignment(Value: TAlignment);
begin
  if (cvAlignment In FAssignedValues) and (Value = FAlignment) then Exit;
  FAlignment := Value;
  Include(FAssignedValues, cvAlignment);
  GetGrid.InvalidateRow(Index);
  Changed(False);
end;

procedure TRow.SetButtonStyle(Value: TColumnButtonStyle);
begin
  if Value = FButtonStyle then Exit;
  FButtonStyle := Value;
  GetGrid.InvalidateRow(Index);
  Changed(False);
end;

procedure TRow.SetPickList(Value: TStrings);
begin
  if Value = nil then
  begin
      FPickList.Free;
      FPickList := nil;
      Exit;
  end;
  PickList.Assign(Value);
end;

procedure TRow.SetPopupMenu(Value: TPopupMenu);
begin
  FPopupMenu := Value;
  if Value <> nil then Value.FreeNotification(GetGrid);
end;

procedure TRow.SetReadOnly(Value: Boolean);
begin
  if (cvReadOnly In FAssignedValues) and (Value = FReadOnly) then Exit;
  FReadOnly := Value;
  Include(FAssignedValues, cvReadOnly);
  Changed(False);
end;

function TRow.GetEditMask: String;
begin
  Result := FEditMask;
  if FEditMask = '' then begin
    if Assigned(Field) then
      Result := Field.EditMask;
    if Result = '' then
      if Assigned(FOnGetEditMask) then FOnGetEditMask(Grid, Grid.EditRow, Result);
  end;
end;

procedure TRow.EditButtonClick;
begin
  if Assigned(FOnEditButtonClick) then
    FOnEditButtonClick(GetGrid)
  else
    if Assigned(Grid.OnEditButtonClick) then
      Grid.OnEditButtonClick(Grid, Index, Field);
end;

function TRow.DefaultImeMode: TImeMode;
var
  Grid: TrRecordGrid;
begin
  Grid := GetGrid;
  if Assigned(Grid) then
    Result := Grid.ImeMode
  else
    Result := FImeMode;
end;

function TRow.DefaultImeName: TImeName;
var
  Grid: TrRecordGrid;
begin
  Grid := GetGrid;
  if Assigned(Grid) then
    Result := Grid.ImeName
  else
    Result := FImeName;
end;

function TRow.GetImeMode: TImeMode;
begin
  if cvImeMode In FAssignedValues then
    Result := FImeMode
  else
    Result := DefaultImeMode;
end;

function TRow.GetImeName: TImeName;
begin
  if cvImeName In FAssignedValues then
    Result := FImeName
  else
    Result := DefaultImeName;
end;

function TRow.IsImeModeStored: Boolean;
begin
  Result := (cvImeMode In FAssignedValues) and (FImeMode <> DefaultImeMode);
end;

function TRow.IsImeNameStored: Boolean;
begin
  Result := (cvImeName In FAssignedValues) and (FImeName <> DefaultImeName);
end;

procedure TRow.SetImeMode(Value: TImeMode);
begin
  if (cvImeMode In FAssignedValues) Or (Value <> DefaultImeMode) then
  begin
    FImeMode := Value;
    Include(FAssignedValues, cvImeMode);
  end;
  Changed(False);
end;

procedure TRow.SetImeName(Value: TImeName);
begin
  if (cvImeName in FAssignedValues) or (Value <> DefaultImeName) then
  begin
    FImeName := Value;
    Include(FAssignedValues, cvImeName);
  end;
  Changed(False);
end;

function TRow.GetText: String;
begin
  Result := FValue;
end;

procedure TRow.SetText(Value: String);
begin
  FValue := Value;
  GetGrid.InvalidateRow(Index);
  FieldName := '';
end;

procedure TRow.Refresh;
begin
  GetGrid.InvalidateRow(Index);
end;

constructor TRow.Create(Collection: TCollection);
begin
  inherited Create(Collection);
  FTitle := CreateTitle;
  FFont := TFont.Create;
  FFont.Assign(DefaultFont);
  FFont.OnChange := FontChanged;
  GetGrid.RowCount := Collection.Count;
  FDropDownRows := 7;
  FButtonStyle := cbsAuto;
  FImeMode := imDontCare;
  FImeName := Screen.DefaultIme;
  FObject := nil;
  FParent := nil;
  FNodes := TRecGridRows.Create(Grid, TRow, False);
  FIndentLevel := 0;
  FExpanded := False;
  FHasChildren := False;
  FIDCode := '';
  FStored := True;
  GetGrid.State := gsCustomized;
end;

destructor TRow.Destroy;
begin
  Grid.RowCount := Collection.Count - 1;
  FFont.Free;
  FPickList.Free;
  FNodes.Free;
  FTitle.Free;
  inherited Destroy;
end;

procedure TRow.SetTitle(Value: TRowTitle);
begin
  FTitle.Assign(Value);
end;

function TRow.CreateTitle: TRowTitle;
begin
  Result := TRowTitle.Create(Self);
end;

function TRow.DefaultFont: TFont;
var
  Grid: TrRecordGrid;
begin
  Grid := GetGrid;
  if Assigned(Grid) then
    Result := Grid.Font
  else
    Result := FFont;
end;

function TRow.GetFont: TFont;
var
  Save: TNotifyEvent;
begin
  if not (cvFont in FAssignedValues) and (FFont.Handle <> DefaultFont.Handle) then
  begin
    Save := FFont.OnChange;
    FFont.OnChange := nil;
    FFont.Assign(DefaultFont);
    FFont.OnChange := Save;
  end;
  Result := FFont;
end;

function TRow.IsFontStored: Boolean;
begin
  Result := (cvFont in FAssignedValues);
end;

procedure TRow.RefreshDefaultFont;
var
  Save: TNotifyEvent;
begin
  if cvFont in FAssignedValues then Exit;
  Save := FFont.OnChange;
  FFont.OnChange := nil;
  try
    FFont.Assign(DefaultFont);
  finally
    FFont.OnChange := Save;
  end;
end;

procedure TRow.SetFont(Value: TFont);
begin
  FFont.Assign(Value);
  Include(FAssignedValues, cvFont);
  Changed(False);
end;

procedure TRow.FontChanged(Sender: TObject);
begin
  Include(FAssignedValues, cvFont);
  Title.RefreshDefaultFont;
  Changed(False);
end;

{ TRecGridRows }

function TRecGridRows.GetColByID(Index: Variant): TRow;
var
  nCnt, nSub: Integer;
begin
  Result := nil;
  Index := AnsiUpperCase(Index);
  for nCnt := 0 to Count - 1 do begin
    if AnsiUpperCase(TRow(Items[nCnt]).FIDCode) = Index then begin
      Result := TRow(Items[nCnt]);
      Break;
    end else
    if TRow(Items[nCnt]).Nodes.Count > 0 then
      for nSub := 0 to TRow(Items[nCnt]).Nodes.Count - 1 do
        if AnsiUpperCase(TRow(TRow(Items[nCnt]).Nodes.Items[nSub]).FIDCode) = Index then begin
          Result := TRow(TRow(Items[nCnt]).Nodes.Items[nSub]);
          Break;
        end
  end;
end;

function TRecGridRows.GetColumn(Index: Variant): TRow;
begin
  if Vartype(Index) = VarString
    then Result := GetColByID(Index)
    else
      if Index >= 0
        then Result := TRow(inherited Items[Index])
        else Result := nil;
end;

procedure TRecGridRows.SetColumn(Index: Variant; Value: TRow);
begin
  if vartype(Index) = varString
    then GetColByID(Index).Assign(Value)
    else if Index >= 0
      then Items[Index].Assign(Value);
end;

function TRecGridRows.GetOwner: TPersistent;
begin
  Result := FGrid;
end;

function TRecGridRows.Add: TRow;
begin
//  if not FEmpty then
    Result := TRow(inherited Add)
{  else
  begin
    FEmpty := False;
    Result := Items[0];
  end;}
end;

procedure TRecGridRows.Clear;
begin
//  FEmpty := False;
  inherited Clear;
  //Add;
  //FEmpty := True;
end;

procedure TRecGridRows.Delete(Index: Variant);
begin
  Items[Index].Free;
  if Count = 0 then Clear;
end;

constructor TRecGridRows.Create(Grid: TrRecordGrid; ColumnClass: TColumnClass; AddEmpty: Boolean);
begin
  inherited Create(ColumnClass);
  FGrid := Grid;
  if AddEmpty then Add;
//  FEmpty := True;
end;

procedure TRecGridRows.RebuildRows;

  procedure AddFields(Fields: TFields; Depth: Integer);
  var
    I: Integer;
  begin
    Inc(Depth);
    for I := 0 to Fields.Count-1 do
    begin
      if Fields[i].Visible then begin
        Add.FieldName := Fields[I].FullName;
        if Fields[I].DataType in [ftADT, ftArray] then
          AddFields((Fields[I] as TObjectField).Fields, Depth);
      end;
    end;
  end;

begin
  if Assigned(FGrid) and Assigned(FGrid.DataSource) and
    Assigned(FGrid.Datasource.Dataset) then
  begin
//    FGrid.BeginUpDate;
    try
      Clear;
      AddFields(FGrid.Datasource.Dataset.Fields, 0);
      FGrid.State := gsDefault;
    finally
//      FGrid.EndUpDate;
    end
  end
  else
    Clear;
end;

{ TrGridDataLink }

procedure TrGridDataLink.CheckBrowseMode;
begin
  if DataSet.State in [dsEdit,dsInsert] then
    if FGrid.Editor <> nil then
      if FGrid.Editor.Modified then begin
        FGrid.Editor.Modified := False;
        TrRecGridInplaceEdit(FGrid.Editor).CloseUp(True);
      end;
end;

procedure TrGridDataLink.ActiveChanged;
begin
  if Active then begin
    FGrid.UpdateEdit;
    if (FGrid.Rows.Count = 0) and FGrid.FRebuildRowsIfEmpty then
      FGrid.Rows.RebuildRows
    else
  end
  else begin
    if FGrid.State = gsDefault then
      FGrid.Rows.Clear;
    FGrid.ResyncFields;
  end;
  FGrid.Invalidate;
end;

procedure TrGridDataLink.DataSetScrolled(Distance: Integer);
begin
  FGrid.InUpdating := False;
  FGrid.UpdateEdit;
  FGrid.Invalidate;
end;

procedure TrGridDataLink.RecordChanged(Field: TField);
var
  nCnt: Integer;
begin
  if DataSet.State = dsBrowse
    then FGrid.InUpdating := False;
  if Field = nil then begin
    FGrid.Invalidate;
    if not FGrid.InUpdating
      then FGrid.UpdateEdit;
  end else
    with FGrid do
      if not InUpdating then
        for nCnt := 0 to Rows.Count - 1 do
          if Rows[nCnt].Field = Field then begin
            InvalidateRow(nCnt);
            if nCnt = EditRow then UpdateEdit;
          end;
end;

constructor TrGridDataLink.Create(AGrid: TrRecordGrid);
begin
  inherited Create;
  FGrid := AGrid;
end;

destructor TrGridDataLink.Destroy;
begin
  inherited Destroy;
end;

{ TPopupListBox }

procedure TPopupListBox.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  with Params do
  begin
    Style := Style Or WS_BORDER;
    ExStyle := WS_EX_TOOLWINDOW Or WS_EX_TOPMOST;
    WindowClass.Style := CS_SAVEBITS;
  end;
end;

procedure TPopupListbox.CreateWnd;
begin
  inherited CreateWnd;
  Windows.SetParent(Handle, 0);
  CallWindowProc(DefWndProc, Handle, wm_SetFocus, 0, 0);
end;

procedure TPopupListbox.KeyPress(var Key: Char);
var
  TickCount: Integer;
begin
  case Key Of
    #8, #27: FSearchText := '';
    #32..#255 :
      begin
        TickCount := GetTickCount;
        if TickCount - FSearchTickCount > 2000 then FSearchText := '';
        FSearchTickCount := TickCount;
        if Length(FSearchText) < 32 then FSearchText := FSearchText + Key;
        SendMessage(Handle, LB_SelectString, WORD(-1), Longint(PChar(FSearchText)));
        Key := #0;
      end;
  end;
  inherited Keypress(Key);
end;

procedure TPopupListbox.MouseUp(Button: TMouseButton; Shift: TShiftState;
    X, Y: Integer);
begin
  inherited MouseUp(Button, Shift, X, Y);
  TrRecGridInplaceEdit(Owner).CloseUp((X >= 0) and (Y >= 0) and
      (X < Width) and (Y < Height));
end;

{ TrRecGridInplaceEdit }

{
procedure TrRecGridInplaceEdit.SetFocus;
begin
  If IsWindowVisible(Handle) and
     (TrRecordGrid(Grid). then Windows.SetFocus(Handle);
end;
}

constructor TrRecGridInplaceEdit.Create(Owner: TComponent);
begin
  inherited Create(Owner);
  FLookupSource := TDataSource.Create(Self);
  FButtonWidth := GetSystemMetrics(SM_CXVSCROLL);
  FEditStyle := esSimple;
  ParentCtl3D := False;
  Ctl3D := False;
  TabStop := False;
  BorderStyle := bsNone;
  FSavedCancel := nil;
end;

procedure TrRecGridInplaceEdit.CatchCancel(Dataset: TDataSet);
begin
  Reset;
  if Assigned(FSavedCancel) then FSavedCancel(DataSet);
end;


procedure TrRecGridInplaceEdit.BoundsChanged;
var
  R: TRect;
begin
  SetRect(R, 2, 2, Width - 2, Height);
  if FEditStyle <> esSimple then Dec(R.Right, FButtonWidth);
  SendMessage(Handle, EM_SETRECTNP, 0, LongInt(@R));
  SendMessage(Handle, EM_SCROLLCARET, 0, 0);
  if SysLocale.Fareast then
      SetImeCompositionWindow(Font, R.Left, R.Top);
end;

procedure TrRecGridInplaceEdit.CloseUp(Accept: Boolean);
var
  MasterField: TField;
  ListValue: Variant;
  nCnt: Integer;
  nIdx: Integer;
  //RSV
  CRow: TRow;

begin
  if FListVisible then
  begin
    if GetCapture <> 0 then SendMessage(GetCapture, WM_CANCELMODE, 0, 0);
    if FActiveList = FDataList then
      ListValue := FDataList.KeyValue
    else
      if FPickList.ItemIndex <> -1 then
          ListValue := FPickList.Items[FPicklist.ItemIndex];

    SetWindowPos(FActiveList.Handle, 0, 0, 0, 0, 0, SWP_NOZORDER Or
        SWP_NOMOVE Or SWP_NOSIZE Or SWP_NOACTIVATE Or SWP_HIDEWINDOW);
    FListVisible := False;
    if Assigned(FDataList) then
        FDataList.ListSource := nil;
    FLookupSource.Dataset := nil;
    Invalidate;
    if Accept then
      if FActiveList = FDataList then
        with TrRecordGrid(Grid), Rows[EditRow].Field do
        begin
          MasterField := DataSet.FieldByName(KeyFields);
          if MasterField.CanModify then
          begin
            if Assigned(GetColumn.OnValidate) then
              GetColumn.OnValidate(Grid, EditRow, Rows[EditRow].PickList, ListValue, Accept);

            if Accept then
            begin
              InUpdating := True;
//RSV
              if not (DataSet.State in [dsEdit, dsInsert]) then
              DataSet.Edit;
              MasterField.Value := ListValue;
              UpdateContents;
              SelectAll;
              Accept := False;
            end
            else
              Reset;
          end;
        end
      else
        if (not varIsNull(ListValue)) and EditCanModify then
          Text := ListValue;
  end;

  if Accept and EditCanModify then
  begin
    if FEditStyle in [esPickList, esDropDown] then
    begin
      nIdx := PickList.Items.IndexOf(Self.Text);

      if nIdx = -1 then
        for nCnt := 0 to PickList.Items.Count-1 do
            if AnsiUpperCase(Self.Text) = AnsiUpperCase(PickList.Items.Strings[nCnt]) then
        begin
          Self.Text := PickList.Items.Strings[nCnt];
          nIdx := nCnt;
          Break;
        end;
      if (FEditStyle = esPickList) and (nIdx = -1) then
      begin
        Reset;
        Exit;
      end;
    end;

    ListValue := Text;

//RSV    if Assigned(GetColumn.OnValidate) then
    CRow := GetColumn;
    if Assigned(CRow) and Assigned(CRow.OnValidate) then
      with TrRecordGrid(Grid) do
          GetColumn.OnValidate(Grid, EditRow, CRow.PickList, ListValue, Accept);

    if Accept then
    begin
      Text := ListValue;
      UpdateField;
    end
    else
      Reset;
  end;
end;

procedure TrRecGridInplaceEdit.UpdateField;
var
  S: String;
begin
    with TrRecordGrid(Grid), Rows[EditRow] do
    if Assigned(Rows[EditRow]) then //RSV new
//RSV
    if not Assigned(Field) or (Assigned(Field) and not IsSpecialField(Field)) then
    begin
        if Assigned(Field) then
        begin
            if DataSource.DataSet.Active then
            begin
//??????????????????
//                if DataSource.DataSet.FieldByName(FieldName).Text <> Self.Text then
                if DataSource.DataSet.FieldByName(FieldName).AsString <> Self.Text then
                begin
                    InUpdating := True;
//RSV
                    if Assigned(DataSource.DataSet) and not (DataSource.DataSet.State in [dsEdit, dsInsert]) then
                    DataSource.DataSet.Edit;
                    S := DataSource.DataSet.FieldByName(FieldName).AsString;
                    try
                      if DataSource.DataSet.FieldByName(FieldName).Datatype = ftMemo then
                        DataSource.DataSet.FieldByName(FieldName).AsString := Self.Text
                      else
                        DataSource.DataSet.FieldByName(FieldName).Text := Self.Text;
                    finally
                      Self.Text := S;
                    end;
//                    DataSource.DataSet.FieldByName(FieldName).AsString := Self.Text;
                    InUpdating := False;
                end;
            end;
        end
        else
            if not (vartype(Value) = varEmpty) then
                Value := Self.Text;

        if Assigned(OnSetText) then OnSetText(TrRecordGrid(Grid), EditRow, Modified, PickList, Self.Text);
    end;

    UpdateContents;
    SelectAll;
end;

procedure TrRecGridInplaceEdit.DoDropDownKeys(var Key: Word; Shift: TShiftState);
begin
  case Key Of
    VK_UP, VK_DOWN :
      if ssAlt In Shift then
      begin
        if FListVisible then CloseUp(True) else DropDown;
        Key := 0;
      end;
    VK_RETURN, VK_ESCAPE:
      if FListVisible and not (ssAlt In Shift) then
      begin
        CloseUp(Key = VK_RETURN);
        Key := 0;
      end;
  end;
end;

procedure TrRecGridInplaceEdit.DropDown;
var
  P: TPoint;
  I, J, Y: Integer;
  Column: TRow;
begin
  if not FListVisible and Assigned(FActiveList) then
  begin
    FActiveList.Width := Width;
    with TrRecordGrid(Grid) do
        Column := Rows[Row];
    if FActiveList = FDataList then
      with Column.Field do
      begin
        FDataList.Color := Color;
        FDataList.Font := Font;
        FDataList.RowCount := Column.DropDownRows;
        FLookupSource.DataSet := LookupDataSet;
        FDataList.KeyField := LookupKeyFields;
        FDataList.ListField := LookupResultField;
        FDataList.ListSource := FLookupSource;
        FDataList.KeyValue := DataSet.FieldByName(KeyFields).Value;
      end
    else
    begin
      FPickList.Color := Color;
      FPickList.Font := Font;
      FPickList.Items := Column.Picklist;
      if Cardinal(FPickList.Items.Count) >= Column.DropDownRows then
        FPickList.Height := FPickList.ItemHeight * Integer(Column.DropDownRows) + 4
      else
        FPickList.Height := FPickList.Items.Count * FPickList.ItemHeight + 4;

      if Assigned(Column.Field) then
      begin
        if Column.Field.IsNull then
          FPickList.ItemIndex := -1
        else
          FPickList.ItemIndex := FPickList.Items.IndexOf(Column.Field.Value);
      end
      else
        FPickList.ItemIndex := FPickList.Items.IndexOf(Text);

      J := FPickList.ClientWidth;
      for I := 0 to FPickList.Items.Count - 1 do
      begin
        Y := FPickList.Canvas.TextWidth(FPickList.Items[I]);
        if Y > J then J := Y;
      end;
      FPickList.ClientWidth := J;
    end;
    P := Parent.ClientToScreen(Point(Left, Top));
    Y := P.Y + Height;
    if Y + FActiveList.Height > Screen.Height then Y := P.Y - FActiveList.Height;
    SetWindowPos(FActiveList.Handle, HWND_TOP, P.X, Y, 0, 0,
      SWP_NOSIZE Or SWP_NOACTIVATE Or SWP_SHOWWINDOW);
    FListVisible := True;
    Invalidate;
    Windows.SetFocus(Handle);
  end;
end;

type
    TWinControlCracker = class(TWinControl) end;

function TrRecGridInplaceEdit.GetColumn: TRow;
begin
  with TrRecordGrid(Grid) do
      Result := Rows[EditRow];
end;

procedure TrRecGridInplaceEdit.KeyDown(var Key: Word; Shift: TShiftState);
begin
  if (ssCtrl in Shift) and (Key in [VK_SUBTRACT, VK_ADD]) and Assigned(Grid)
  then TrRecordGrid(Grid).KeyDown(Key,Shift);

  if (EditStyle = esEllipsis) and (Key = VK_RETURN) and (Shift = [ssCtrl]) then
  begin
    GetColumn.EditButtonClick;
    KillMessage(Handle, WM_CHAR);
  end
  else
    inherited KeyDown(Key, Shift);
end;

procedure TrRecGridInplaceEdit.ListMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  if Button = mbLeft then
    CloseUp(PtInRect(FActiveList.ClientRect, Point(X, Y)));
end;

procedure TrRecGridInplaceEdit.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
  if (Button = mbLeft) and (FEditStyle <> esSimple) and
    PtInRect(Rect(Width - FButtonWidth, 0, Width, Height), Point(X, Y)) then
  begin
    if FListVisible then
      CloseUp(False)
    else
    begin
      MouseCapture := True;
      FTracking := True;
      TrackButton(X, Y);
      if Assigned(FActiveList) then
          DropDown;
    end;
  end;
  inherited MouseDown(Button, Shift, X, Y);
end;

procedure TrRecGridInplaceEdit.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  ListPos: TPoint;
  MousePos: TSmallPoint;
begin
  if FTracking then
  begin
    TrackButton(X, Y);
    if FListVisible then
    begin
      ListPos := FActiveList.ScreenToClient(ClientToScreen(Point(X, Y)));
      if PtInRect(FActiveList.ClientRect, ListPos) then
      begin
        StopTracking;
        MousePos := PointToSmallPoint(ListPos);
        SendMessage(FActiveList.Handle, WM_LBUTTONDOWN, 0, Integer(MousePos));
        Exit;
      end;
    end;
  end;
  inherited MouseMove(Shift, X, Y);
end;

procedure TrRecGridInplaceEdit.MouseUp(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
var
  WasPressed: Boolean;
begin
  WasPressed := FPressed;
  StopTracking;
  if (Button = mbLeft) and (FEditStyle = esEllipsis) and WasPressed then
    GetColumn.EditButtonClick;
  inherited MouseUp(Button, Shift, X, Y);
end;

procedure TrRecGridInplaceEdit.PaintWindow(DC: HDC);
var
  R: TRect;
  Flags: Integer;
  W: Integer;
begin
  if FEditStyle <> esSimple then
  begin
    SetRect(R, Width - FButtonWidth, 0, Width, Height);
    Flags := 0;
    if FEditStyle In [esDataList, esPickList, esDropDown] then
    begin
      if FActiveList = nil then
        Flags := DFCS_INACTIVE
      else if FPressed then
        Flags := DFCS_FLAT Or DFCS_PUSHED;
      DrawFrameControl(DC, R, DFC_SCROLL, Flags Or DFCS_SCROLLCOMBOBOX);
    end
    else { esEllipsis }
    begin
      if FPressed then
          Flags := BF_FLAT;
      DrawEdge(DC, R, EDGE_RAISED, BF_RECT Or BF_MIDDLE Or Flags);
      Flags := ((R.Right - R.Left) Shr 1) - 1 + Ord(FPressed);
      W := Height Shr 3;
      if W = 0 then W := 1;
      PatBlt(DC, R.Left + Flags, R.Top + Flags, W, W, BLACKNESS);
      PatBlt(DC, R.Left + Flags - (W * 2), R.Top + Flags, W, W, BLACKNESS);
      PatBlt(DC, R.Left + Flags + (W * 2), R.Top + Flags, W, W, BLACKNESS);
    end;
    ExcludeClipRect(DC, R.Left, R.Top, R.Right, R.Bottom);
  end;
  inherited PaintWindow(DC);
end;

procedure TrRecGridInplaceEdit.SetEditStyle(Value: TEditStyle);
begin
  if Value = FEditStyle then Exit;
  FEditStyle := Value;
  case Value Of
    esPickList, esDropDown :
      begin
        if FPickList = nil then
        begin
          FPickList := TPopupListbox.Create(Self);
          FPickList.Visible := False;
          FPickList.Parent := Self;
          FPickList.OnMouseUp := ListMouseUp;
          FPickList.IntegralHeight := True;
          FPickList.ItemHeight := 11;
        end;
        FActiveList := FPickList;
      end;
    esDataList :
      begin
        if FDataList = nil then
        begin
          FDataList := TPopupDataList.Create(Self);
          FDataList.Visible := False;
          FDataList.Parent := Self;
          FDataList.OnMouseUp := ListMouseUp;
        end;
        FActiveList := FDataList;
      end;
  else { cbsNone, cbsEllipsis, or read only field }
    FActiveList := nil;
  end;
  with TrRecordGrid(Grid) do
    Self.ReadOnly := Rows[Row].ReadOnly;
  Repaint;
end;

procedure TrRecGridInplaceEdit.StopTracking;
begin
  if FTracking then
  begin
    TrackButton(-1, -1);
    FTracking := False;
    MouseCapture := False;
  end;
end;

procedure TrRecGridInplaceEdit.TrackButton(X, Y: Integer);
var
  NewState: Boolean;
  R: TRect;
begin
  SetRect(R, ClientWidth - FButtonWidth, 0, ClientWidth, ClientHeight);
  NewState := PtInRect(R, Point(X, Y));
  if FPressed <> NewState then
  begin
    FPressed := NewState;
    InvalidateRect(Handle, @R, False);
  end;
end;

procedure TrRecGridInplaceEdit.UpdateContents;
var
  Column: TRow;
  NewStyle: TEditStyle;
  MasterField: TField;
begin
  with TrRecordGrid(Grid) do
      Column := Rows[Row];

  Self.ReadOnly := Column.ReadOnly;
  NewStyle := esSimple;

  case Column.ButtonStyle Of
    cbsEllipsis: NewStyle := esEllipsis;
    cbsAuto: begin
        if Assigned(Column.Field) then
        begin
          with Column.Field do
          begin
            { Show the dropdown button only if the field is editable }
            if FieldKind = fkLookup then
            begin
              MasterField := Dataset.FieldByName(KeyFields);
              { Column.DefaultReadonly will always be True for a lookup field.
               Test if Column.ReadOnly has been assigned a value of True }
              if Assigned(MasterField) and
                MasterField.CanModify and not
                ((cvReadOnly In Column.AssignedValues) and
                Column.ReadOnly) then
                with TrRecordGrid(Grid) do
                    if not ReadOnly and
                        DataLink.Active and not
                        Datalink.ReadOnly then
                        NewStyle := esDataList
            end
            else
            begin
              if IsSpecialField(Column.Field) then
              begin
                if Assigned(TrRecordGrid(Grid).OnEditButtonClick) then
                    NewStyle := esEllipsis;
              end;
            end;
          end;
        end;

        if Assigned(Column.Picklist) and
          (Column.PickList.Count > 0) and not Column.Readonly then
          NewStyle := esPickList;
      end;
    cbsDropDown    : if Assigned(Column.Picklist) and
                        (Column.PickList.Count > 0) and not Column.Readonly then
                        NewStyle := esDropDown;
    cbsDropDownList: if Assigned(Column.Picklist) and
                        (Column.PickList.Count > 0) and not Column.Readonly then
                        NewStyle := esPickList;
  end;
  EditStyle := NewStyle;
  inherited UpdateContents;
end;

procedure TrRecGridInplaceEdit.CMCancelMode(var Message: TCMCancelMode);
begin
  //!!!!!!!!!!!!!!!!!!
  if (Message.Sender <> Self) and (Message.Sender <> FActiveList) then
      CloseUp(False);
end;

procedure TrRecGridInplaceEdit.WMCancelMode(var Message: TMessage);
begin
  StopTracking;
  inherited;
end;

procedure TrRecGridInplaceEdit.WMSetFocus(var Msg: TWMSetFocus);
begin
  inherited;
  if (TrRecordGrid(Grid).DataSource <> nil) then
    if (TrRecordGrid(Grid).DataSource.DataSet <> nil) and not Assigned(FSavedCancel) then
  begin
    FSavedCancel := TrRecordGrid(Owner).DataSource.DataSet.BeforeCancel;
    TrRecordGrid(Owner).DataSource.DataSet.BeforeCancel := Self.CatchCancel;
  end;
end;

procedure TrRecGridInplaceEdit.WMKillFocus(var Message: TMessage);
begin
  if SysLocale.FarEast then
  begin
    ImeName := Screen.DefaultIme;
    ImeMode := imDontCare;
  end;
  inherited;
  CloseUp(Modified);
  if TrRecordGrid(Owner).DataSource <> nil then
    if TrRecordGrid(Owner).Datasource.DataSet <> nil then
      TrRecordGrid(Owner).DataSource.DataSet.BeforeCancel := FSavedCancel;
  FSavedCancel := nil;
end;

procedure TrRecGridInplaceEdit.WMLButtonDblClk(var Message: TWMLButtonDblClk);
var
  nIdx: Integer;
  oStr: TStrings;
  nCnt: Integer;
begin
  with Message do
    if (FEditStyle <> esSimple) then
    begin
      if PtInRect(Rect(Width - FButtonWidth, 0, Width, Height), Point(XPos, YPos)) then Exit;

      if FEditStyle in [esPickList, esDropDown] then
      begin
        with TrRecordGrid(Grid) do FPickList.Items := Rows[Row].PickList;
        if FPickList.Items.Count > 0 then
          oStr := FPickList.Items
        else
          with TrRecordGrid(Grid) do oStr := Rows[Row].PickList;
        if oStr.Count > 0 then
        begin
          nIdx := oStr.IndexOf(Self.Text);
          if nIdx = -1 then
            for nCnt := 0 to oStr.Count-1 do
              if AnsiUpperCase(Self.Text) = AnsiUpperCase(oStr.Strings[nCnt]) then
            begin
              Self.Text := oStr.Strings[nCnt];
              nIdx := nCnt;
              Break;
            end;
          if nIdx <> -1 then
          begin
            if nIdx < oStr.Count - 1 then
              Self.Text := oStr.Strings[nIdx + 1]
            else
              Self.Text := oStr.Strings[0];
          end
          else
            Self.Text := oStr.Strings[0];
          Modified := True;
          with GetColumn, TrRecordGrid(Grid) do
          begin
              if Assigned(OnSetEditText) then OnSetEditText(Self, Row, PickList, Self.Text);
          end;
          CloseUp(True);
        end;
      end;
    end;
  inherited;
end;

procedure TrRecGridInplaceEdit.WMPaint(var Message: TWMPaint);
begin
  PaintHandler(Message);
end;

procedure TrRecGridInplaceEdit.WMSetCursor(var Message: TWMSetCursor);
var P: TPoint;
begin
  GetCursorPos(P);
  if (FEditStyle <> esSimple) and
    PtInRect(Rect(Width - FButtonWidth, 0, Width, Height), ScreenToClient(P)) then
    Windows.SetCursor(LoadCursor(0, idc_Arrow))
  else
    inherited;
end;

procedure TrRecGridInplaceEdit.WndProc(var Message: TMessage);
begin
  case Message.Msg of
    wm_KeyDown, wm_SysKeyDown, wm_Char :
      if EditStyle In [esPickList,esDropDown, esDataList] then
        with TWMKey(Message) do
        begin
          doDropDownKeys(CharCode, KeyDataToShiftState(KeyData));
          if (CharCode <> 0) and FListVisible then
          begin
            with TMessage(Message) do
                SendMessage(FActiveList.Handle, Msg, WParam, LParam);
            Exit;
          end;
        end
  end;
  inherited;
end;

{ TRowTitle }

constructor TRowTitle.Create(Column: TRow);
begin
  inherited Create;
  FColumn := Column;
  FFont := TFont.Create;
  FFont.Assign(DefaultFont);
  FFont.OnChange := FontChanged;
end;

destructor TRowTitle.Destroy;
begin
  FFont.Free;
  inherited Destroy;
end;

procedure TRowTitle.Assign(Source: TPersistent);
begin
  if Source is TRowTitle then
  begin
    if cvTitleAlignment in TRowTitle(Source).FColumn.FAssignedValues then
      Alignment := TRowTitle(Source).Alignment;
    if cvTitleColor in TRowTitle(Source).FColumn.FAssignedValues then
      Color := TRowTitle(Source).Color;
    if cvTitleCaption in TRowTitle(Source).FColumn.FAssignedValues then
      Caption := TRowTitle(Source).Caption;
    if cvTitleFont in TRowTitle(Source).FColumn.FAssignedValues then
      Font := TRowTitle(Source).Font;
  end
  else
    inherited Assign(Source);
end;

function TRowTitle.DefaultAlignment: TAlignment;
begin
  Result := taLeftJustify;
end;

function TRowTitle.DefaultColor: TColor;
var
  Grid: TrRecordGrid;
begin
  Grid := FColumn.GetGrid;
  if Assigned(Grid) then
    Result := Grid.FixedColor
  else
    Result := clBtnFace;
end;

function TRowTitle.DefaultFont: TFont;
var
  Grid: TrRecordGrid;
begin
  Grid := FColumn.GetGrid;
  if Assigned(Grid) then
    Result := Grid.Font
  else
    Result := FColumn.Font;
end;

function TRowTitle.DefaultCaption: string;
var
  Field: TField;
begin
  Field := FColumn.Field;
  if Assigned(Field) then
    Result := Field.DisplayName
  else
    Result := FColumn.FieldName;
end;

procedure TRowTitle.FontChanged(Sender: TObject);
begin
  Include(FColumn.FAssignedValues, cvTitleFont);
  FColumn.Changed(True);
end;

function TRowTitle.GetAlignment: TAlignment;
begin
  if cvTitleAlignment in FColumn.FAssignedValues then
    Result := FAlignment
  else
    Result := DefaultAlignment;
end;

function TRowTitle.GetColor: TColor;
begin
  if cvTitleColor in FColumn.FAssignedValues then
    Result := FColor
  else
    Result := DefaultColor;
end;

function TRowTitle.GetCaption: string;
begin
  if cvTitleCaption in FColumn.FAssignedValues then
    Result := FCaption
  else
    Result := DefaultCaption;
end;

function TRowTitle.GetFont: TFont;
var
  Save: TNotifyEvent;
  Def: TFont;
begin
  if not (cvTitleFont in FColumn.FAssignedValues) then
  begin
    Def := DefaultFont;
    if (FFont.Handle <> Def.Handle) or (FFont.Color <> Def.Color) then
    begin
      Save := FFont.OnChange;
      FFont.OnChange := nil;
      FFont.Assign(DefaultFont);
      FFont.OnChange := Save;
    end;
  end;
  Result := FFont;
end;

function TRowTitle.IsAlignmentStored: Boolean;
begin
  Result := (cvTitleAlignment in FColumn.FAssignedValues) and
    (FAlignment <> DefaultAlignment);
end;

function TRowTitle.IsColorStored: Boolean;
begin
  Result := (cvTitleColor in FColumn.FAssignedValues) and
    (FColor <> DefaultColor);
end;

function TRowTitle.IsFontStored: Boolean;
begin
  Result := (cvTitleFont in FColumn.FAssignedValues);
end;

function TRowTitle.IsCaptionStored: Boolean;
begin
  Result := (cvTitleCaption in FColumn.FAssignedValues) and
    (FCaption <> DefaultCaption);
end;

procedure TRowTitle.RefreshDefaultFont;
var
  Save: TNotifyEvent;
begin
  if (cvTitleFont in FColumn.FAssignedValues) then Exit;
  Save := FFont.OnChange;
  FFont.OnChange := nil;
  try
    FFont.Assign(DefaultFont);
  finally
    FFont.OnChange := Save;
  end;
end;

procedure TRowTitle.RestoreDefaults;
var
  FontAssigned: Boolean;
begin
  FontAssigned := cvTitleFont in FColumn.FAssignedValues;
  FColumn.FAssignedValues := FColumn.FAssignedValues - ColumnTitleValues;
  FCaption := '';
  RefreshDefaultFont;
  FColumn.Changed(FontAssigned);
end;

procedure TRowTitle.SetAlignment(Value: TAlignment);
begin
  if (cvTitleAlignment in FColumn.FAssignedValues) and (Value = FAlignment) then Exit;
  FAlignment := Value;
  Include(FColumn.FAssignedValues, cvTitleAlignment);
  FColumn.Changed(False);
  with FColumn do GetGrid.InvalidateRow(Index);
end;

procedure TRowTitle.SetColor(Value: TColor);
begin
  if (cvTitleColor in FColumn.FAssignedValues) and (Value = FColor) then Exit;
  FColor := Value;
  Include(FColumn.FAssignedValues, cvTitleColor);
  FColumn.Changed(False);
  with FColumn do GetGrid.InvalidateRow(Index);
end;

procedure TRowTitle.SetFont(Value: TFont);
begin
  FFont.Assign(Value);
  with FColumn do GetGrid.InvalidateRow(Index);
end;

procedure TRowTitle.SetCaption(const Value: string);
var
 Grid: TrRecordGrid;
begin
  if Column.IsStored then
  begin
    if (cvTitleCaption in FColumn.FAssignedValues) and (Value = FCaption) then Exit;
    FCaption := Value;
    Include(Column.FAssignedValues, cvTitleCaption);
    Column.Changed(False);
  end
  else
  begin
    Grid := Column.GetGrid;
    if Assigned(Grid) and (Grid.Datalink.Active) and Assigned(Column.Field) then
      Column.Field.DisplayLabel := Value;
  end;
  with FColumn do GetGrid.InvalidateRow(Index);
end;

end.

